import importlib
from django.conf import settings


def get_settings():


    return getattr(settings, "CRYPTOCURRENCY_PAYMENT", {})


def get_backend_config(crypto, key=None):
    crypto_backend = get_settings().get(crypto.upper())
    if not crypto_backend or crypto_backend["ACTIVE"] is not True:
        raise Exception("{} backend not found".format(crypto))
    if key:
        return crypto_backend[key]
    return crypto_backend


def get_active_backends():
    return [key for key, value in get_settings().items() if value["ACTIVE"] is True]


def get_backend_obj(crypto):
    crypto_backend = get_backend_config(crypto=crypto, key="BACKEND")
    module = ".".join(crypto_backend.split(".")[:-1])
    Class = crypto_backend.split(".")[-1]
    module = importlib.import_module(module, Class)
    Backend = getattr(module, Class)
    return Backend(get_backend_config(crypto, key="MASTER_PUBLIC_KEY"))



# -*- coding: utf-8 -*-

from django.contrib import admin

from .models import (
   CryptoCurrencyPayment
)


@admin.register(CryptoCurrencyPayment)
class CrptoCurrancyPaymentAdmin(admin.ModelAdmin):
    pass






from django.db import models

# Put your test models here
from django.db import models
from django.contrib.contenttypes.fields import GenericRelation

from cryptocurrency_payment.models import CryptoCurrencyPayment


class Invoice(models.Model):

    payments = GenericRelation(CryptoCurrencyPayment)
    title = models.CharField(max_length=50)



# -*- coding: utf-8 -*-
from uuid import uuid4

from django.db import models
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType

from cryptocurrency_payment.app_settings import get_backend_config, get_backend_obj


def create_child_payment(payment, fiat_amount):
    """
    Create a child payment from a particular payment. A child payment can be created for a particular underpaid amount
    :param payment: The parent payment
    :param amount: The new fiat_amount. fiat_currency is the same as parent fiat_currency
    :return: payment oj
    """
    child_payment = create_new_payment(
        payment.crypto,
        fiat_amount=fiat_amount,
        fiat_currency=payment.fiat_currency,
        payment_title=payment.payment_title,
        payment_description=payment.payment_description,
        related_object=payment.content_object,
        user=payment.user,
        parent_payment=payment,
    )
    payment.child_payment = child_payment
    payment.save()
    return child_payment


def create_new_payment(
    crypto,
    fiat_amount,
    fiat_currency,
    payment_title=None,
    payment_description=None,
    related_object=None,
    user=None,
    parent_payment=None,
    address_index=None,
    reuse_address=None,
):
    """
    Create a cryptocurrency payment by using this method instead of creating directly on through the main model
    It helps with generation of addresses and fait conversion. Call these method directly to create a payment
    :param crypto: The crypto in config this transaction belongs too
    :param fiat_amount: The fiat amount of this transaction
    :param fiat_currency: The fiat currency of this transaction
    :param payment_title: Title you want on the payment
    :param payment_description: Description of the payment
    :param related_object: A generic object of another model that can be linked to this payments payment can be accessed from the model with object.payments.
    :param user: User that owns the payment for non anonymous payment
    :param parent_payment: The parent payment for subpayment or underpaid payments
    :param address_index: Use a particular generated address index for address generation see merchant-wallet
    :return: payment obj
    """
    crypto_reuse_address = reuse_address or get_backend_config(
        crypto, key="REUSE_ADDRESS"
    )
    crypto_code = get_backend_config(crypto, key="CODE")
    backend_obj = get_backend_obj(crypto)

    crypto_amount = backend_obj.convert_from_fiat(fiat_amount, fiat_currency)
    address = None
    resuse_address = False
    related_object_id = None
    if related_object:
        related_object_id = related_object.pk
    if parent_payment:
        address = parent_payment.address
        resuse_address = True
    if not address and crypto_reuse_address is True:
        address = CryptoCurrencyPayment.get_crypto_reused_address(crypto)
        resuse_address = address is not None
    if not address:
        address_generated_count = (
            address_index or CryptoCurrencyPayment.get_address_used_count(crypto)
        )
        address = backend_obj.generate_new_address(index=address_generated_count)
    payment = CryptoCurrencyPayment(
        crypto=crypto,
        crypto_code=crypto_code,
        address=address,
        crypto_amount=crypto_amount,
        fiat_amount=fiat_amount,
        fiat_currency=fiat_currency,
        payment_title=payment_title,
        payment_description=payment_description,
        address_reused=resuse_address,
        user=user,
        object_id=related_object_id,
        content_object=related_object,
        parent_payment=parent_payment,
    )
    if fiat_amount == 0:
        payment.status = CryptoCurrencyPayment.PAYMENT_PAID
    payment.save()
    return payment


class CryptoCurrencyPayment(models.Model):
    """
    Cryptocurrencypayment model to handle all cryptocurrency transactions and transaction status
    underpaid balance payment are recorded as child payment with their parent linked
    """

    PAYMENT_PAID = "paid"
    PAYMENT_CANCELLED = "cancelled"
    PAYMENT_NEW = "new"
    PAYMENT_PROCESSING = "processing"

    TRANSACTION_STATUS = (
        (PAYMENT_NEW, "New"),
        (PAYMENT_PAID, "Paid"),
        (PAYMENT_CANCELLED, "Cancelled"),
        (PAYMENT_PROCESSING, "Processing"),
    )
    id = models.UUIDField(
        default=uuid4, editable=False, primary_key=True, verbose_name="ID",
    )
    crypto = models.CharField(max_length=50)
    crypto_code = models.CharField(max_length=15, null=True)
    address = models.CharField(max_length=200)
    address_reused = models.BooleanField(default=False)
    tx_hash = models.CharField(max_length=150, null=True, help_text="Transaction ID")

    status = models.CharField(
        choices=TRANSACTION_STATUS, default=PAYMENT_NEW, max_length=50
    )

    crypto_amount = models.DecimalField(
        max_digits=17, decimal_places=12, help_text="Calculated crypto amount for fiat"
    )
    paid_crypto_amount = models.DecimalField(
        max_digits=17,
        decimal_places=12,
        help_text="Paid crypto amount",
        null=True,
        default=0,
    )
    fiat_amount = models.DecimalField(
        max_digits=9, decimal_places=2, help_text="Fiat amount"
    )
    fiat_currency = models.CharField(max_length=5, default="USD")
    payment_title = models.CharField(max_length=100, null=True)
    payment_description = models.CharField(max_length=250, null=True)
    user = models.ForeignKey(
        get_user_model(),
        null=True,
        on_delete=models.CASCADE,
        related_name="crypto_payments",
    )
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True)
    object_id = models.CharField(null=True, max_length=100)
    content_object = GenericForeignKey("content_type", "object_id")
    parent_payment = models.OneToOneField(
        "self", null=True, on_delete=models.CASCADE, related_name="child"
    )
    child_payment = models.OneToOneField(
        "self", null=True, on_delete=models.SET_NULL, related_name="parent"
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "{} {} {} {}".format(
            self.crypto_amount, self.address, self.fiat_amount, self.fiat_currency
        )

    @classmethod
    def get_crypto_reused_address(cls, crypto):
        """
        Get an address from old paid transaction when reusing address
        :param crypto: The crypto to get its address
        :return:
        """

        previous_payment = (
            cls.objects.filter(crypto=crypto, status=cls.PAYMENT_PAID)
            .order_by("updated_at")
            .first()
        )
        if previous_payment:
            address = previous_payment.address
            return address

    @classmethod
    def get_address_used_count(cls, crypto):
        """
        Get the total number of addresses generated for a particluar crypto currency
        :param crypto:
        :return:
        """
        return cls.objects.filter(crypto=crypto).count()

    @property
    def remaining_crypto_amount(self):
        if self.crypto_amount > self.paid_crypto_amount:
            return self.crypto_amount - self.paid_crypto_amount

    @property
    def is_under_paid(self):
        if self.paid_crypto_amount and self.remaining_crypto_amount:
            return True
        return False



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-05-13 15:11
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CryptoCurrencyPayment',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name=b'ID')),
                ('crypto', models.CharField(max_length=50)),
                ('crypto_code', models.CharField(max_length=15, null=True)),
                ('address', models.CharField(max_length=200)),
                ('address_reused', models.BooleanField(default=False)),
                ('tx_hash', models.CharField(help_text=b'Transaction ID', max_length=150, null=True)),
                ('status', models.CharField(choices=[(b'new', b'New'), (b'paid', b'Paid'), (b'cancelled', b'Cancelled'), (b'processing', b'Processing')], default=b'new', max_length=50)),
                ('crypto_amount', models.DecimalField(decimal_places=12, help_text=b'Calculated crypto amount for fiat', max_digits=17)),
                ('paid_crypto_amount', models.DecimalField(decimal_places=12, default=0, help_text=b'Paid crypto amount', max_digits=17, null=True)),
                ('fiat_amount', models.DecimalField(decimal_places=2, help_text=b'Fiat amount', max_digits=9)),
                ('fiat_currency', models.CharField(default=b'USD', max_length=5)),
                ('payment_title', models.CharField(max_length=100, null=True)),
                ('payment_description', models.CharField(max_length=250, null=True)),
                ('object_id', models.CharField(max_length=100, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('child_payment', models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='parent', to='cryptocurrency_payment.CryptoCurrencyPayment')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('parent_payment', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='child', to='cryptocurrency_payment.CryptoCurrencyPayment')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='crypto_payments', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from django.views.generic import DetailView
from django.http import Http404
from cryptocurrency_payment.models import CryptoCurrencyPayment
from cryptocurrency_payment.app_settings import get_backend_config


class CryptoPaymentDetailView(DetailView):
    queryset = CryptoCurrencyPayment.objects.all()
    context_object_name = 'payment'
    template_name = 'cryptocurrency_payment/payment_detail.html'

    def get_object(self, *args ):

        obj = super(CryptoPaymentDetailView, self).get_object( *args )
        allow_anon_payment = get_backend_config(obj.crypto, key='ALLOW_ANONYMOUS_PAYMENT')
        if  allow_anon_payment is not True and   self.request.user.is_authenticated is not True:
            raise Http404
        elif   obj.user and self.request.user != obj.user:
            raise Http404
        return obj



# -*- coding: utf-8 -*-
from django.conf.urls import url

from cryptocurrency_payment import views

app_name = 'cryptocurrency_payment'
urlpatterns = [

    url(
        regex="^payment/(?P<pk>[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12})/$",
        view=views.CryptoPaymentDetailView.as_view(),
        name='crypto_payment_detail',
    ),

]



from cryptocurrency_payment.models import CryptoCurrencyPayment
from django.utils import timezone
from cryptocurrency_payment.models import create_child_payment
from datetime import timedelta
from cryptocurrency_payment.app_settings import get_active_backends, get_backend_config, get_backend_obj


def update_payment_status():
    """
    Run this as a task periodically to check update for new payment and current processing payment on the blockchain
    :return:
    """
    backends = get_active_backends()
    for backend in backends:
        crypto_task = CryptoCurrencyPaymentTask(backend)
        crypto_task.update_crypto_currency_payment_status()


def cancel_unpaid_payment():
    """
    Run this as a task to cancel payment that have stayed in new for too long
    :return:
    """
    backends = get_active_backends()
    for backend in backends:
        crypto_task = CryptoCurrencyPaymentTask(backend)
        crypto_task.cancel_unpaid_payment()


def refresh_payment_prices():
    """
    Payment prices can be renewed periodically according to the latest conversion rate using this method
    :return:
    """
    backends = get_active_backends()
    for backend in backends:
        crypto_task = CryptoCurrencyPaymentTask(backend)
        crypto_task.refresh_new_crypto_payment_amount()


class CryptoCurrencyPaymentTask:
    """
    Implements task for a particular crypto backend . You can cancel a unpaid payment,
    Update unpaid payment status and refresh unpaid payment prices
    """

    def __init__(self, crypto):

        self.unpaid_payment_hrs = get_backend_config(
            crypto, "CANCEL_UNPAID_PAYMENT_HRS"
        )
        self.crypto = crypto
        self.backend_obj = get_backend_obj(crypto)
        self.ignore_underpayment_amount = get_backend_config(
            crypto, "IGNORE_UNDERPAYMENT_AMOUNT"
        )
        self.create_new_underpayment = get_backend_config(
            crypto, "CREATE_NEW_UNDERPAID_PAYMENT"
        )
        self.refresh_prices_every_mins = get_backend_config(
            crypto, "REFRESH_PRICE_AFTER_MINUTE"
        )
        self.confirmation_number = get_backend_config(
            crypto, "BALANCE_CONFIRMATION_NUM"
        )
        self.confirm_bal_without_hash_mins = get_backend_config(
            crypto, "IGNORE_CONFIRMED_BALANCE_WITHOUT_SAVED_HASH_MINS"
        )

    def update_crypto_currency_payment_status(self):
        """
        Get all payment that are in new status or processing status and check their status on
        the blockchain for confirmation. Only payment that are still in this particular status
        are checked

        :return:
        """
        yesterday_time = timezone.now() - timedelta(hours=self.unpaid_payment_hrs)
        payments = CryptoCurrencyPayment.objects.filter(
            crypto=self.crypto,
            status__in=[
                CryptoCurrencyPayment.PAYMENT_NEW,
                CryptoCurrencyPayment.PAYMENT_PROCESSING,
            ],
            created_at__gte=yesterday_time,
        ).all()
        for payment in payments:
            status, value = self.backend_obj.confirm_address_payment(
                address=payment.address,
                total_crypto_amount=payment.crypto_amount,
                confirmation_number=self.confirmation_number,
                accept_confirmed_bal_without_hash_mins=self.confirm_bal_without_hash_mins,
                tx_hash=payment.tx_hash,
            )
            if status == self.backend_obj.UNCONFIRMED_ADDRESS_BALANCE:
                payment.status = payment.PAYMENT_PROCESSING
                payment.tx_hash = value
            elif status == self.backend_obj.CONFIRMED_ADDRESS_BALANCE:
                payment.status = payment.PAYMENT_PAID
                payment.paid_crypto_amount = value
            elif status == self.backend_obj.UNDERPAID_ADDRESS_BALANCE:
                fiat_value = self.backend_obj.convert_to_fiat(
                    value, payment.fiat_currency
                )
                if (
                    self.create_new_underpayment
                    and fiat_value > self.ignore_underpayment_amount
                ):
                    payment.child_payment = create_child_payment(payment, fiat_value)
                payment.status = payment.PAYMENT_PAID
                payment.paid_crypto_amount = value
            else:
                # unknown error occured cancel payment
                payment.status = payment.PAYMENT_CANCELLED
            payment.save()

    def cancel_unpaid_payment(self):
        """
        Any unpaid payment still in new payment status less than a particular time can be cancelled
        . To reduce resources when checking for new payment status
        :return:
        """
        yesterday_time = timezone.now() - timedelta(hours=self.unpaid_payment_hrs)
        payments = CryptoCurrencyPayment.objects.filter(
            status__in=[CryptoCurrencyPayment.PAYMENT_NEW],
            created_at__lte=yesterday_time,
        )
        for payment in payments:
            payment.status = payment.PAYMENT_CANCELLED
            payment.save()

    def refresh_new_crypto_payment_amount(self):
        """
        Due to volatility of crypto prices, Payment prices can be refreshed regularly especially for payment in
        new status
        :return:
        """
        leastupdate_time = timezone.now() - timedelta(
            minutes=self.refresh_prices_every_mins
        )
        payments = CryptoCurrencyPayment.objects.filter(
            status=CryptoCurrencyPayment.PAYMENT_NEW, updated_at__lte=leastupdate_time
        )
        for payment in payments:
            amount = self.backend_obj.convert_from_fiat(
                payment.fiat_amount, payment.fiat_currency
            )
            payment.crypto_amount = amount
            payment.save()
